@@ -1,12 +1,12 @@
 import sys
 
+import pystencils_autodiff._backport
 import pystencils_autodiff.backends  # NOQA
-from pystencils_autodiff._field_to_tensors import (  # NOQA
-    tf_constant_from_field, tf_placeholder_from_field, tf_scalar_variable_from_field,
-    tf_variable_from_field, torch_tensor_from_field)
+import pystencils_autodiff.tensorflow_jit
 from pystencils_autodiff._adjoint_field import AdjointField
 from pystencils_autodiff._autodiff import (
-    AutoDiffAstPair, AutoDiffOp, create_backward_assignments, get_jacobian_of_assignments, DiffModes)
+    AutoDiffAstPair, AutoDiffBoundaryHandling, AutoDiffOp, DiffModes, create_backward_assignments,
+    get_jacobian_of_assignments)
 
 __all__ = ['backends',
            'AdjointField',
@@ -14,9 +14,9 @@ __all__ = ['backends',
            'create_backward_assignments',
            'AutoDiffOp',
            'AutoDiffAstPair',
-           "tf_constant_from_field", " tf_placeholder_from_field",
-           "tf_scalar_variable_from_field", " tf_variable_from_field",
-           "torch_tensor_from_field", "DiffModes"]
+           'tensorflow_jit',
+           'DiffModes',
+           'AutoDiffBoundaryHandling']
 
 sys.modules['pystencils.autodiff'] = pystencils_autodiff
 sys.modules['pystencils.autodiff.backends'] = pystencils_autodiff.backends
