@@ -6,4 +6,4 @@ a Torch or a Tensorflow operation or we can compile a static
 library to be directly loaded into Torch/Tensorflow.
 """
 
-AVAILABLE_BACKENDS = ['tensorflow', 'torch', 'tensorflow_cpp', 'torch_native']
+AVAILABLE_BACKENDS = ['tensorflow', 'torch', 'tensorflow_native', 'torch_native']
